"""
Checkpoint engine module for SGLang.

This module provides functionality for updating model weights via checkpoint engine.
"""

from sglang.srt.checkpoint_engine.update import main

__all__ = ["main"]
