# Copied and adapted from: https://github.com/hao-ai-lab/FastVideo

from sglang.multimodal_gen.configs.sample.diffusers_generic import (
    DiffusersGenericSamplingParams,
)
from sglang.multimodal_gen.configs.sample.sampling_params import SamplingParams

__all__ = ["SamplingParams", "DiffusersGenericSamplingParams"]
