# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: sglang_scheduler.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'sglang_scheduler.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16sglang_scheduler.proto\x12\x15sglang.grpc.scheduler\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xd0\x05\n\x0eSamplingParams\x12\x13\n\x0btemperature\x18\x01 \x01(\x02\x12\r\n\x05top_p\x18\x02 \x01(\x02\x12\r\n\x05top_k\x18\x03 \x01(\x05\x12\r\n\x05min_p\x18\x04 \x01(\x02\x12\x19\n\x11\x66requency_penalty\x18\x05 \x01(\x02\x12\x18\n\x10presence_penalty\x18\x06 \x01(\x02\x12\x1a\n\x12repetition_penalty\x18\x07 \x01(\x02\x12\x1b\n\x0emax_new_tokens\x18\x08 \x01(\x05H\x01\x88\x01\x01\x12\x0c\n\x04stop\x18\t \x03(\t\x12\x16\n\x0estop_token_ids\x18\n \x03(\r\x12\x1b\n\x13skip_special_tokens\x18\x0b \x01(\x08\x12%\n\x1dspaces_between_special_tokens\x18\x0c \x01(\x08\x12\x0f\n\x05regex\x18\r \x01(\tH\x00\x12\x15\n\x0bjson_schema\x18\x0e \x01(\tH\x00\x12\x16\n\x0c\x65\x62nf_grammar\x18\x0f \x01(\tH\x00\x12\x18\n\x0estructural_tag\x18\x10 \x01(\tH\x00\x12\t\n\x01n\x18\x11 \x01(\x05\x12\x16\n\x0emin_new_tokens\x18\x12 \x01(\x05\x12\x12\n\nignore_eos\x18\x13 \x01(\x08\x12\x14\n\x0cno_stop_trim\x18\x14 \x01(\x08\x12\x1c\n\x0fstream_interval\x18\x15 \x01(\x05H\x02\x88\x01\x01\x12H\n\nlogit_bias\x18\x16 \x03(\x0b\x32\x34.sglang.grpc.scheduler.SamplingParams.LogitBiasEntry\x12.\n\rcustom_params\x18\x17 \x01(\x0b\x32\x17.google.protobuf.Struct\x1a\x30\n\x0eLogitBiasEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x42\x0c\n\nconstraintB\x11\n\x0f_max_new_tokensB\x12\n\x10_stream_interval\"]\n\x13\x44isaggregatedParams\x12\x16\n\x0e\x62ootstrap_host\x18\x01 \x01(\t\x12\x16\n\x0e\x62ootstrap_port\x18\x02 \x01(\x05\x12\x16\n\x0e\x62ootstrap_room\x18\x03 \x01(\x05\"\xe2\x04\n\x0fGenerateRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x38\n\ttokenized\x18\x02 \x01(\x0b\x32%.sglang.grpc.scheduler.TokenizedInput\x12:\n\tmm_inputs\x18\x03 \x01(\x0b\x32\'.sglang.grpc.scheduler.MultimodalInputs\x12>\n\x0fsampling_params\x18\x04 \x01(\x0b\x32%.sglang.grpc.scheduler.SamplingParams\x12\x16\n\x0ereturn_logprob\x18\x05 \x01(\x08\x12\x19\n\x11logprob_start_len\x18\x06 \x01(\x05\x12\x18\n\x10top_logprobs_num\x18\x07 \x01(\x05\x12\x19\n\x11token_ids_logprob\x18\x08 \x03(\r\x12\x1c\n\x14return_hidden_states\x18\t \x01(\x08\x12H\n\x14\x64isaggregated_params\x18\n \x01(\x0b\x32*.sglang.grpc.scheduler.DisaggregatedParams\x12\x1e\n\x16\x63ustom_logit_processor\x18\x0b \x01(\t\x12-\n\ttimestamp\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0blog_metrics\x18\r \x01(\x08\x12\x14\n\x0cinput_embeds\x18\x0e \x03(\x02\x12\x0f\n\x07lora_id\x18\x0f \x01(\t\x12\x1a\n\x12\x64\x61ta_parallel_rank\x18\x10 \x01(\x05\x12\x0e\n\x06stream\x18\x11 \x01(\x08\":\n\x0eTokenizedInput\x12\x15\n\roriginal_text\x18\x01 \x01(\t\x12\x11\n\tinput_ids\x18\x02 \x03(\r\"\xd3\x01\n\x10MultimodalInputs\x12\x12\n\nimage_urls\x18\x01 \x03(\t\x12\x12\n\nvideo_urls\x18\x02 \x03(\t\x12\x12\n\naudio_urls\x18\x03 \x03(\t\x12\x33\n\x12processed_features\x18\x04 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x12\n\nimage_data\x18\x05 \x03(\x0c\x12\x12\n\nvideo_data\x18\x06 \x03(\x0c\x12\x12\n\naudio_data\x18\x07 \x03(\x0c\x12\x12\n\nmodalities\x18\x08 \x03(\t\"\xe3\x01\n\x10GenerateResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12;\n\x05\x63hunk\x18\x02 \x01(\x0b\x32*.sglang.grpc.scheduler.GenerateStreamChunkH\x00\x12;\n\x08\x63omplete\x18\x03 \x01(\x0b\x32\'.sglang.grpc.scheduler.GenerateCompleteH\x00\x12\x35\n\x05\x65rror\x18\x04 \x01(\x0b\x32$.sglang.grpc.scheduler.GenerateErrorH\x00\x42\n\n\x08response\"\x95\x02\n\x13GenerateStreamChunk\x12\x11\n\ttoken_ids\x18\x01 \x03(\r\x12\x15\n\rprompt_tokens\x18\x02 \x01(\x05\x12\x19\n\x11\x63ompletion_tokens\x18\x03 \x01(\x05\x12\x15\n\rcached_tokens\x18\x04 \x01(\x05\x12>\n\x0foutput_logprobs\x18\x05 \x01(\x0b\x32%.sglang.grpc.scheduler.OutputLogProbs\x12\x15\n\rhidden_states\x18\x06 \x03(\x02\x12<\n\x0einput_logprobs\x18\x07 \x01(\x0b\x32$.sglang.grpc.scheduler.InputLogProbs\x12\r\n\x05index\x18\x08 \x01(\r\"\x9b\x03\n\x10GenerateComplete\x12\x12\n\noutput_ids\x18\x01 \x03(\r\x12\x15\n\rfinish_reason\x18\x02 \x01(\t\x12\x15\n\rprompt_tokens\x18\x03 \x01(\x05\x12\x19\n\x11\x63ompletion_tokens\x18\x04 \x01(\x05\x12\x15\n\rcached_tokens\x18\x05 \x01(\x05\x12>\n\x0foutput_logprobs\x18\x06 \x01(\x0b\x32%.sglang.grpc.scheduler.OutputLogProbs\x12>\n\x11\x61ll_hidden_states\x18\x07 \x03(\x0b\x32#.sglang.grpc.scheduler.HiddenStates\x12\x1a\n\x10matched_token_id\x18\x08 \x01(\rH\x00\x12\x1a\n\x10matched_stop_str\x18\t \x01(\tH\x00\x12<\n\x0einput_logprobs\x18\n \x01(\x0b\x32$.sglang.grpc.scheduler.InputLogProbs\x12\r\n\x05index\x18\x0b \x01(\rB\x0e\n\x0cmatched_stop\"K\n\rGenerateError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x18\n\x10http_status_code\x18\x02 \x01(\t\x12\x0f\n\x07\x64\x65tails\x18\x03 \x01(\t\"u\n\x0eOutputLogProbs\x12\x16\n\x0etoken_logprobs\x18\x01 \x03(\x02\x12\x11\n\ttoken_ids\x18\x02 \x03(\x05\x12\x38\n\x0ctop_logprobs\x18\x03 \x03(\x0b\x32\".sglang.grpc.scheduler.TopLogProbs\"\x9e\x01\n\rInputLogProbs\x12@\n\x0etoken_logprobs\x18\x01 \x03(\x0b\x32(.sglang.grpc.scheduler.InputTokenLogProb\x12\x11\n\ttoken_ids\x18\x02 \x03(\x05\x12\x38\n\x0ctop_logprobs\x18\x03 \x03(\x0b\x32\".sglang.grpc.scheduler.TopLogProbs\"1\n\x11InputTokenLogProb\x12\x12\n\x05value\x18\x01 \x01(\x02H\x00\x88\x01\x01\x42\x08\n\x06_value\"0\n\x0bTopLogProbs\x12\x0e\n\x06values\x18\x01 \x03(\x02\x12\x11\n\ttoken_ids\x18\x02 \x03(\x05\"?\n\x0cHiddenStates\x12\x0e\n\x06values\x18\x01 \x03(\x02\x12\r\n\x05layer\x18\x02 \x01(\x05\x12\x10\n\x08position\x18\x03 \x01(\x05\"\xca\x02\n\x0c\x45mbedRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x38\n\ttokenized\x18\x02 \x01(\x0b\x32%.sglang.grpc.scheduler.TokenizedInput\x12:\n\tmm_inputs\x18\x04 \x01(\x0b\x32\'.sglang.grpc.scheduler.MultimodalInputs\x12>\n\x0fsampling_params\x18\x05 \x01(\x0b\x32%.sglang.grpc.scheduler.SamplingParams\x12\x13\n\x0blog_metrics\x18\x06 \x01(\x08\x12\x16\n\x0etoken_type_ids\x18\x07 \x03(\x05\x12\x1a\n\x12\x64\x61ta_parallel_rank\x18\x08 \x01(\x05\x12\x18\n\x10is_cross_encoder\x18\t \x01(\x08\x12\r\n\x05texts\x18\n \x03(\t\"\x9d\x01\n\rEmbedResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x38\n\x08\x63omplete\x18\x02 \x01(\x0b\x32$.sglang.grpc.scheduler.EmbedCompleteH\x00\x12\x32\n\x05\x65rror\x18\x03 \x01(\x0b\x32!.sglang.grpc.scheduler.EmbedErrorH\x00\x42\n\n\x08response\"\xa3\x01\n\rEmbedComplete\x12\x11\n\tembedding\x18\x01 \x03(\x02\x12\x15\n\rprompt_tokens\x18\x02 \x01(\x05\x12\x15\n\rcached_tokens\x18\x03 \x01(\x05\x12\x15\n\rembedding_dim\x18\x04 \x01(\x05\x12:\n\x10\x62\x61tch_embeddings\x18\x05 \x03(\x0b\x32 .sglang.grpc.scheduler.Embedding\"*\n\tEmbedding\x12\x0e\n\x06values\x18\x01 \x03(\x02\x12\r\n\x05index\x18\x02 \x01(\x05\"<\n\nEmbedError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x0f\n\x07\x64\x65tails\x18\x03 \x01(\t\"\x14\n\x12HealthCheckRequest\"7\n\x13HealthCheckResponse\x12\x0f\n\x07healthy\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"2\n\x0c\x41\x62ortRequest\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\"1\n\rAbortResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"I\n\x0fLoadLoRARequest\x12\x12\n\nadapter_id\x18\x01 \x01(\t\x12\x14\n\x0c\x61\x64\x61pter_path\x18\x02 \x01(\t\x12\x0c\n\x04rank\x18\x03 \x01(\x05\"H\n\x10LoadLoRAResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x12\n\nadapter_id\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\"\'\n\x11UnloadLoRARequest\x12\x12\n\nadapter_id\x18\x01 \x01(\t\"6\n\x12UnloadLoRAResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"w\n\x14UpdateWeightsRequest\x12\x13\n\tdisk_path\x18\x01 \x01(\tH\x00\x12\x15\n\x0btensor_data\x18\x02 \x01(\x0cH\x00\x12\x14\n\nremote_url\x18\x03 \x01(\tH\x00\x12\x13\n\x0bweight_name\x18\x04 \x01(\tB\x08\n\x06source\"9\n\x15UpdateWeightsResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"-\n\x17GetInternalStateRequest\x12\x12\n\nstate_keys\x18\x01 \x03(\t\"B\n\x18GetInternalStateResponse\x12&\n\x05state\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"A\n\x17SetInternalStateRequest\x12&\n\x05state\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\"<\n\x18SetInternalStateResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"\x15\n\x13GetModelInfoRequest\"\xac\x03\n\x14GetModelInfoResponse\x12\x12\n\nmodel_path\x18\x01 \x01(\t\x12\x16\n\x0etokenizer_path\x18\x02 \x01(\t\x12\x15\n\ris_generation\x18\x03 \x01(\x08\x12!\n\x19preferred_sampling_params\x18\x04 \x01(\t\x12\x16\n\x0eweight_version\x18\x05 \x01(\t\x12\x19\n\x11served_model_name\x18\x06 \x01(\t\x12\x1a\n\x12max_context_length\x18\x07 \x01(\x05\x12\x12\n\nvocab_size\x18\x08 \x01(\x05\x12\x17\n\x0fsupports_vision\x18\t \x01(\x08\x12\x12\n\nmodel_type\x18\n \x01(\t\x12\x15\n\reos_token_ids\x18\x0b \x03(\x05\x12\x14\n\x0cpad_token_id\x18\x0c \x01(\x05\x12\x14\n\x0c\x62os_token_id\x18\r \x01(\x05\x12\x19\n\x11max_req_input_len\x18\x0e \x01(\x05\x12\x15\n\rarchitectures\x18\x0f \x03(\t\x12\x15\n\rid2label_json\x18\x10 \x01(\t\x12\x12\n\nnum_labels\x18\x11 \x01(\x05\"\x16\n\x14GetServerInfoRequest\"\xb7\x02\n\x15GetServerInfoResponse\x12,\n\x0bserver_args\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12/\n\x0escheduler_info\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x17\n\x0f\x61\x63tive_requests\x18\x03 \x01(\x05\x12\x11\n\tis_paused\x18\x04 \x01(\x08\x12\x1e\n\x16last_receive_timestamp\x18\x05 \x01(\x01\x12\x16\n\x0euptime_seconds\x18\x06 \x01(\x01\x12\x16\n\x0esglang_version\x18\x07 \x01(\t\x12\x13\n\x0bserver_type\x18\x08 \x01(\t\x12.\n\nstart_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"D\n\x0fGetLoadsRequest\x12\x14\n\x07\x64p_rank\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x0f\n\x07include\x18\x02 \x03(\tB\n\n\x08_dp_rank\"\xbe\x01\n\x10GetLoadsResponse\x12\x11\n\ttimestamp\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x15\n\rdp_rank_count\x18\x03 \x01(\x05\x12\x33\n\x05loads\x18\x04 \x03(\x0b\x32$.sglang.grpc.scheduler.SchedulerLoad\x12:\n\taggregate\x18\x05 \x01(\x0b\x32\'.sglang.grpc.scheduler.AggregateMetrics\"\x99\x05\n\rSchedulerLoad\x12\x0f\n\x07\x64p_rank\x18\x01 \x01(\x05\x12\x18\n\x10num_running_reqs\x18\x02 \x01(\x05\x12\x18\n\x10num_waiting_reqs\x18\x03 \x01(\x05\x12\x16\n\x0enum_total_reqs\x18\x04 \x01(\x05\x12\x17\n\x0fnum_used_tokens\x18\x05 \x01(\x05\x12\x1c\n\x14max_total_num_tokens\x18\x06 \x01(\x05\x12\x13\n\x0btoken_usage\x18\x07 \x01(\x01\x12\x16\n\x0egen_throughput\x18\x08 \x01(\x01\x12\x16\n\x0e\x63\x61\x63he_hit_rate\x18\t \x01(\x01\x12\x13\n\x0butilization\x18\n \x01(\x01\x12\x1c\n\x14max_running_requests\x18\x0b \x01(\x05\x12\x39\n\x06memory\x18\x0c \x01(\x0b\x32$.sglang.grpc.scheduler.MemoryMetricsH\x00\x88\x01\x01\x12\x43\n\x0bspeculative\x18\r \x01(\x0b\x32).sglang.grpc.scheduler.SpeculativeMetricsH\x01\x88\x01\x01\x12\x35\n\x04lora\x18\x0e \x01(\x0b\x32\".sglang.grpc.scheduler.LoRAMetricsH\x02\x88\x01\x01\x12I\n\x0e\x64isaggregation\x18\x0f \x01(\x0b\x32,.sglang.grpc.scheduler.DisaggregationMetricsH\x03\x88\x01\x01\x12\x38\n\x06queues\x18\x10 \x01(\x0b\x32#.sglang.grpc.scheduler.QueueMetricsH\x04\x88\x01\x01\x42\t\n\x07_memoryB\x0e\n\x0c_speculativeB\x07\n\x05_loraB\x11\n\x0f_disaggregationB\t\n\x07_queues\"a\n\rMemoryMetrics\x12\x11\n\tweight_gb\x18\x01 \x01(\x01\x12\x13\n\x0bkv_cache_gb\x18\x02 \x01(\x01\x12\x10\n\x08graph_gb\x18\x03 \x01(\x01\x12\x16\n\x0etoken_capacity\x18\x04 \x01(\x05\"@\n\x12SpeculativeMetrics\x12\x15\n\raccept_length\x18\x01 \x01(\x01\x12\x13\n\x0b\x61\x63\x63\x65pt_rate\x18\x02 \x01(\x01\"K\n\x0bLoRAMetrics\x12\x12\n\nslots_used\x18\x01 \x01(\x05\x12\x13\n\x0bslots_total\x18\x02 \x01(\x05\x12\x13\n\x0butilization\x18\x03 \x01(\x01\"\x9c\x02\n\x15\x44isaggregationMetrics\x12\x0c\n\x04mode\x18\x01 \x01(\t\x12#\n\x1bprefill_prealloc_queue_reqs\x18\x02 \x01(\x05\x12#\n\x1bprefill_inflight_queue_reqs\x18\x03 \x01(\x05\x12\"\n\x1a\x64\x65\x63ode_prealloc_queue_reqs\x18\x04 \x01(\x05\x12\"\n\x1a\x64\x65\x63ode_transfer_queue_reqs\x18\x05 \x01(\x05\x12#\n\x1b\x64\x65\x63ode_retracted_queue_reqs\x18\x06 \x01(\x05\x12\x1e\n\x16kv_transfer_speed_gb_s\x18\x07 \x01(\x01\x12\x1e\n\x16kv_transfer_latency_ms\x18\x08 \x01(\x01\"S\n\x0cQueueMetrics\x12\x0f\n\x07waiting\x18\x01 \x01(\x05\x12\x0f\n\x07grammar\x18\x02 \x01(\x05\x12\x0e\n\x06paused\x18\x03 \x01(\x05\x12\x11\n\tretracted\x18\x04 \x01(\x05\"\xa8\x01\n\x10\x41ggregateMetrics\x12\x1a\n\x12total_running_reqs\x18\x01 \x01(\x05\x12\x1a\n\x12total_waiting_reqs\x18\x02 \x01(\x05\x12\x12\n\ntotal_reqs\x18\x03 \x01(\x05\x12\x17\n\x0f\x61vg_token_usage\x18\x04 \x01(\x01\x12\x16\n\x0e\x61vg_throughput\x18\x05 \x01(\x01\x12\x17\n\x0f\x61vg_utilization\x18\x06 \x01(\x01\x32\xb0\x05\n\x0fSglangScheduler\x12]\n\x08Generate\x12&.sglang.grpc.scheduler.GenerateRequest\x1a\'.sglang.grpc.scheduler.GenerateResponse0\x01\x12R\n\x05\x45mbed\x12#.sglang.grpc.scheduler.EmbedRequest\x1a$.sglang.grpc.scheduler.EmbedResponse\x12\x64\n\x0bHealthCheck\x12).sglang.grpc.scheduler.HealthCheckRequest\x1a*.sglang.grpc.scheduler.HealthCheckResponse\x12R\n\x05\x41\x62ort\x12#.sglang.grpc.scheduler.AbortRequest\x1a$.sglang.grpc.scheduler.AbortResponse\x12g\n\x0cGetModelInfo\x12*.sglang.grpc.scheduler.GetModelInfoRequest\x1a+.sglang.grpc.scheduler.GetModelInfoResponse\x12j\n\rGetServerInfo\x12+.sglang.grpc.scheduler.GetServerInfoRequest\x1a,.sglang.grpc.scheduler.GetServerInfoResponse\x12[\n\x08GetLoads\x12&.sglang.grpc.scheduler.GetLoadsRequest\x1a\'.sglang.grpc.scheduler.GetLoadsResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sglang_scheduler_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SAMPLINGPARAMS_LOGITBIASENTRY']._loaded_options = None
  _globals['_SAMPLINGPARAMS_LOGITBIASENTRY']._serialized_options = b'8\001'
  _globals['_SAMPLINGPARAMS']._serialized_start=113
  _globals['_SAMPLINGPARAMS']._serialized_end=833
  _globals['_SAMPLINGPARAMS_LOGITBIASENTRY']._serialized_start=732
  _globals['_SAMPLINGPARAMS_LOGITBIASENTRY']._serialized_end=780
  _globals['_DISAGGREGATEDPARAMS']._serialized_start=835
  _globals['_DISAGGREGATEDPARAMS']._serialized_end=928
  _globals['_GENERATEREQUEST']._serialized_start=931
  _globals['_GENERATEREQUEST']._serialized_end=1541
  _globals['_TOKENIZEDINPUT']._serialized_start=1543
  _globals['_TOKENIZEDINPUT']._serialized_end=1601
  _globals['_MULTIMODALINPUTS']._serialized_start=1604
  _globals['_MULTIMODALINPUTS']._serialized_end=1815
  _globals['_GENERATERESPONSE']._serialized_start=1818
  _globals['_GENERATERESPONSE']._serialized_end=2045
  _globals['_GENERATESTREAMCHUNK']._serialized_start=2048
  _globals['_GENERATESTREAMCHUNK']._serialized_end=2325
  _globals['_GENERATECOMPLETE']._serialized_start=2328
  _globals['_GENERATECOMPLETE']._serialized_end=2739
  _globals['_GENERATEERROR']._serialized_start=2741
  _globals['_GENERATEERROR']._serialized_end=2816
  _globals['_OUTPUTLOGPROBS']._serialized_start=2818
  _globals['_OUTPUTLOGPROBS']._serialized_end=2935
  _globals['_INPUTLOGPROBS']._serialized_start=2938
  _globals['_INPUTLOGPROBS']._serialized_end=3096
  _globals['_INPUTTOKENLOGPROB']._serialized_start=3098
  _globals['_INPUTTOKENLOGPROB']._serialized_end=3147
  _globals['_TOPLOGPROBS']._serialized_start=3149
  _globals['_TOPLOGPROBS']._serialized_end=3197
  _globals['_HIDDENSTATES']._serialized_start=3199
  _globals['_HIDDENSTATES']._serialized_end=3262
  _globals['_EMBEDREQUEST']._serialized_start=3265
  _globals['_EMBEDREQUEST']._serialized_end=3595
  _globals['_EMBEDRESPONSE']._serialized_start=3598
  _globals['_EMBEDRESPONSE']._serialized_end=3755
  _globals['_EMBEDCOMPLETE']._serialized_start=3758
  _globals['_EMBEDCOMPLETE']._serialized_end=3921
  _globals['_EMBEDDING']._serialized_start=3923
  _globals['_EMBEDDING']._serialized_end=3965
  _globals['_EMBEDERROR']._serialized_start=3967
  _globals['_EMBEDERROR']._serialized_end=4027
  _globals['_HEALTHCHECKREQUEST']._serialized_start=4029
  _globals['_HEALTHCHECKREQUEST']._serialized_end=4049
  _globals['_HEALTHCHECKRESPONSE']._serialized_start=4051
  _globals['_HEALTHCHECKRESPONSE']._serialized_end=4106
  _globals['_ABORTREQUEST']._serialized_start=4108
  _globals['_ABORTREQUEST']._serialized_end=4158
  _globals['_ABORTRESPONSE']._serialized_start=4160
  _globals['_ABORTRESPONSE']._serialized_end=4209
  _globals['_LOADLORAREQUEST']._serialized_start=4211
  _globals['_LOADLORAREQUEST']._serialized_end=4284
  _globals['_LOADLORARESPONSE']._serialized_start=4286
  _globals['_LOADLORARESPONSE']._serialized_end=4358
  _globals['_UNLOADLORAREQUEST']._serialized_start=4360
  _globals['_UNLOADLORAREQUEST']._serialized_end=4399
  _globals['_UNLOADLORARESPONSE']._serialized_start=4401
  _globals['_UNLOADLORARESPONSE']._serialized_end=4455
  _globals['_UPDATEWEIGHTSREQUEST']._serialized_start=4457
  _globals['_UPDATEWEIGHTSREQUEST']._serialized_end=4576
  _globals['_UPDATEWEIGHTSRESPONSE']._serialized_start=4578
  _globals['_UPDATEWEIGHTSRESPONSE']._serialized_end=4635
  _globals['_GETINTERNALSTATEREQUEST']._serialized_start=4637
  _globals['_GETINTERNALSTATEREQUEST']._serialized_end=4682
  _globals['_GETINTERNALSTATERESPONSE']._serialized_start=4684
  _globals['_GETINTERNALSTATERESPONSE']._serialized_end=4750
  _globals['_SETINTERNALSTATEREQUEST']._serialized_start=4752
  _globals['_SETINTERNALSTATEREQUEST']._serialized_end=4817
  _globals['_SETINTERNALSTATERESPONSE']._serialized_start=4819
  _globals['_SETINTERNALSTATERESPONSE']._serialized_end=4879
  _globals['_GETMODELINFOREQUEST']._serialized_start=4881
  _globals['_GETMODELINFOREQUEST']._serialized_end=4902
  _globals['_GETMODELINFORESPONSE']._serialized_start=4905
  _globals['_GETMODELINFORESPONSE']._serialized_end=5333
  _globals['_GETSERVERINFOREQUEST']._serialized_start=5335
  _globals['_GETSERVERINFOREQUEST']._serialized_end=5357
  _globals['_GETSERVERINFORESPONSE']._serialized_start=5360
  _globals['_GETSERVERINFORESPONSE']._serialized_end=5671
  _globals['_GETLOADSREQUEST']._serialized_start=5673
  _globals['_GETLOADSREQUEST']._serialized_end=5741
  _globals['_GETLOADSRESPONSE']._serialized_start=5744
  _globals['_GETLOADSRESPONSE']._serialized_end=5934
  _globals['_SCHEDULERLOAD']._serialized_start=5937
  _globals['_SCHEDULERLOAD']._serialized_end=6602
  _globals['_MEMORYMETRICS']._serialized_start=6604
  _globals['_MEMORYMETRICS']._serialized_end=6701
  _globals['_SPECULATIVEMETRICS']._serialized_start=6703
  _globals['_SPECULATIVEMETRICS']._serialized_end=6767
  _globals['_LORAMETRICS']._serialized_start=6769
  _globals['_LORAMETRICS']._serialized_end=6844
  _globals['_DISAGGREGATIONMETRICS']._serialized_start=6847
  _globals['_DISAGGREGATIONMETRICS']._serialized_end=7131
  _globals['_QUEUEMETRICS']._serialized_start=7133
  _globals['_QUEUEMETRICS']._serialized_end=7216
  _globals['_AGGREGATEMETRICS']._serialized_start=7219
  _globals['_AGGREGATEMETRICS']._serialized_end=7387
  _globals['_SGLANGSCHEDULER']._serialized_start=7390
  _globals['_SGLANGSCHEDULER']._serialized_end=8078
# @@protoc_insertion_point(module_scope)
